int output=0;
unsigned short *ramw,*romw;
unsigned short gpuram[0x800];

unsigned long nextpc;
int gpujumpin=0;
int gpurun=0;
int gpuendian;
unsigned long gpupc;
unsigned long gpuflags;
unsigned long gpuhidata;

unsigned readgpu(unsigned addr)
{
        switch (addr)
        {
                case 0xF02114: /*G_CTRL high*/
                return 0;
                case 0xF02116: /*G_CTRL low*/
                if (gpurun) return 1;
                return 0;

                default:
                printf("Bad 68k GPU read %08X\n",addr);
                exit(-1);
        }
}

void writegpu(unsigned addr, unsigned val)
{
        if (addr&0x400)
        switch (addr)
        {
                case 0xF02100: /*G_FLAGS high*/
                gpuflags=(gpuflags&0xFFFF)|(val<<16);
                break;
                case 0xF02102: /*G_FLAGS low*/
                gpuflags=(gpuflags&0xFFFF0000)|val;
                break;

                case 0xF0210C: case 0xF0210E: /*Endian control*/
                gpuendian=val&7;
                break;

                case 0xF02110: /*G_PC high*/
                gpupc=(gpupc&0xFFFF)|(val<<16);
                break;
                case 0xF02112: /*G_PC low*/
                gpupc=(gpupc&0xFFFF0000)|val;
                break;

                case 0xF02114: /*G_CTRL high*/
                break;
                case 0xF02116: /*G_CTRL low*/
                if (val&1) printf("Starting GPU\n");
                else       printf("Stopping GPU\n");
                gpurun=val&1;
                break;

                default:
                printf("Bad 68k GPU write %08X %08X\n",addr,val);
                exit(-1);
        }
}


/*GPU emulation*/
unsigned long gpuregs[2][32];
int gpuset=1;

void dumpgpuregs()
{
        int c;
        printf("PC=%08X FLAGS=%08X\n",gpupc,gpuflags);
        for (c=0;c<8;c++)
            printf("R%02i=%08X R%02i=%08X R%02i=%08X R%02i=%08X\n",c<<2,gpuregs[1][c<<2],(c<<2)+1,gpuregs[1][(c<<2)+1],(c<<2)+2,gpuregs[1][(c<<2)+2],(c<<2)+3,gpuregs[1][(c<<2)+3]);
}

inline unsigned short gquickread(unsigned long addr)
{
        if (addr<0x200000) return ramw[addr>>1];
        if ((addr&0xFFF000)==0xF03000) return gpuram[(addr&0xFFF)>>1];
        printf("Bad GPU read %08X\n",addr);
        exit(-1);
}

inline unsigned short gread(unsigned long addr)
{
        if (addr<0x200000) return ramw[addr>>1];
        if ((addr&0xFFF000)==0xF03000) return gpuram[(addr&0xFFF)>>1];
        printf("Bad GPU read %08X\n",addr);
        exit(-1);
}

inline unsigned short greadb(unsigned long addr)
{
        if (addr<0x200000)
        {
                if (addr&1) return ramw[addr>>1]>>8;
                return ramw[addr>>1];
        }
        if (addr>=0x800000 && addr<0xC00000)
        {
                if (addr&1) return romw[(addr&0x3FFFFF)>>1]>>8;
                return romw[(addr&0x3FFFFF)>>1];
        }
        printf("Bad GPU readb %08X\n",addr);
        exit(-1);
}

inline void gwrite(unsigned long addr, unsigned short val)
{
        if (addr>0xF0000F && addr<0xF00028)
        {
                writeobj(addr,val);
                return;
        }
        if (addr>0xF00400 && addr<0xF007FF)
        {
                writeclut(addr,val);
                return;
        }
        if (addr>=0xF02100 && addr<0xF0211F)
        {
                writegpu(addr,val);
                return;
        }
        if ((addr&0xFFF000)==0xF03000)
        {
                gpuram[(addr&0xFFF)>>1]=val;
                return;
        }
        printf("Bad GPU write %08X %04X\n",addr,val);
        exit(-1);
}

inline void gwriteb(unsigned long addr, unsigned char val)
{
        if (addr<0x200000)
        {
                if (addr&1) ramw[addr>>1]=(ramw[addr>>1]&0xFF)|(val<<8);
                else        ramw[addr>>1]=(ramw[addr>>1]>>8)|val;
                return;
        }
        printf("Bad GPU write %08X %04X\n",addr,val);
        exit(-1);
}

void gpub0rk()
{
        printf("GPU B0rked\n");
        dumpgpuregs();
        exit(-1);
}

#define RN (opcode&31)
#define RM ((opcode>>5)&31)

#define C0 (opcode&0x20)
#define C1 (opcode&0x40)
#define C2 (opcode&0x80)
#define C3 (opcode&0x100)
#define C4 (opcode&0x200)

#define ZFLAG (gpuflags&1)
#define CFLAG (gpuflags&2)
#define NFLAG (gpuflags&4)

inline void setgpuzn(unsigned long a)
{
        if (!a)                gpuflags|=ZFLAG;
        else if (a&0x80000000) gpuflags|=NFLAG;
}

inline void setgpuadd(unsigned long a, unsigned long b)
{
        gpuflags&=~7;
        if (b>(0xFFFFFFFF-a)) gpuflags|=CFLAG;
        a+=b;
        if (!a)                gpuflags|=ZFLAG;
        else if (a&0x80000000) gpuflags|=NFLAG;
}

inline void setgpuadc(unsigned long a, unsigned long b)
{
        int c=gpuflags&1;
        gpuflags&=~7;
        if (b>(0xFFFFFFFF-a)) gpuflags|=CFLAG;
        a+=b+c;
        if (!a)                gpuflags|=ZFLAG;
        else if (a&0x80000000) gpuflags|=NFLAG;
}

inline void setgpusub(unsigned long a, unsigned long b)
{
        gpuflags&=~7;
        if ((a-b)>a) gpuflags|=CFLAG;
        a-=b;
        if (!a)                gpuflags|=ZFLAG;
        else if (a&0x80000000) gpuflags|=NFLAG;
}

void rungpu(int cycles)
{
        unsigned short opcode;
        unsigned long temp;
        while (cycles>0)
        {
                opcode=gquickread(gpupc);
                gpupc+=2;
                switch (opcode>>10)
                {
                        case 0: /*ADD Rm,Rn*/
                        setgpuadd(gpuregs[gpuset][RN],gpuregs[gpuset][RM]);
                        gpuregs[gpuset][RN]+=gpuregs[gpuset][RM];
                        cycles-=3;
                        break;

                        case 1: /*ADDC Rm,Rn*/
                        temp=gpuflags&1;
                        setgpuadc(gpuregs[gpuset][RN],gpuregs[gpuset][RM]);
                        gpuregs[gpuset][RN]+=gpuregs[gpuset][RM]+temp;
                        cycles-=3;
                        break;

                        case 2: /*ADDQ #,Rn*/
                        temp=(RM)?RM:32;
                        setgpuadd(gpuregs[gpuset][RN],temp);
                        gpuregs[gpuset][RN]+=temp;
                        cycles-=3;
                        break;

                        case 4: /*SUB Rm,Rn*/
                        setgpusub(gpuregs[gpuset][RN],gpuregs[gpuset][RM]);
                        gpuregs[gpuset][RN]-=gpuregs[gpuset][RM];
                        cycles-=3;
                        break;

                        case 6: /*SUBQ #,Rn*/
                        temp=(RM)?RM:32;
                        setgpusub(gpuregs[gpuset][RN],temp);
                        gpuregs[gpuset][RN]-=temp;
                        cycles-=3;
                        break;

                        case 7: /*SUBQT #,Rn*/
                        temp=(RM)?RM:32;
                        gpuregs[gpuset][RN]-=temp;
                        cycles-=3;
                        break;

                        case 13: /*BTST #,Rn*/
                        gpuflags&=~6;
                        if (!(gpuregs[gpuset][RN]&(1<<RM))) gpuflags|=ZFLAG;
                        if (gpuregs[gpuset][RN]&0x80000000) gpuflags|=NFLAG;
                        cycles-=3;
                        return;

                        case 14: /*BSET #,Rn*/
                        gpuregs[gpuset][RN]|=(1<<RM);
                        setgpuzn(gpuregs[gpuset][RN]);
                        cycles-=3;
                        return;

                        case 24: /*SHLQ #,Rn*/
                        temp=(RM)?RM:32;
                        gpuflags&=~7;
                        if (gpuregs[gpuset][RN]&0x80000000) gpuflags|=CFLAG;
                        gpuregs[gpuset][RN]<<=(32-temp);
                        setgpuzn(gpuregs[gpuset][RN]);
                        cycles-=3;
                        break;

                        case 29: /*RORQ #,Rn*/
                        temp=(RM)?RM:32;
                        gpuflags&=~7;
                        if (gpuregs[gpuset][RN]&0x80000000) gpuflags|=CFLAG;
                        gpuregs[gpuset][RN]=(gpuregs[gpuset][RN]>>temp)|(gpuregs[gpuset][RN]<<(32-temp));
                        setgpuzn(gpuregs[gpuset][RN]);
                        cycles-=3;
                        break;

                        case 30: /*CMP*/
                        setgpusub(gpuregs[gpuset][RN],gpuregs[gpuset][RM]);
                        cycles-=3;
                        break;

                        case 31: /*CMPQ*/
                        temp=(RM&0x10)?RM|0xFFFFFFF0:RM;
                        setgpusub(gpuregs[gpuset][RN],temp);
                        cycles-=3;
                        break;

                        case 34: /*MOVE Rm,Rn*/
                        gpuregs[gpuset][RN]=gpuregs[gpuset][RM];
                        cycles-=3;
                        break;

                        case 35: /*MOVEQ*/
                        gpuregs[gpuset][RN]=RM;
                        cycles-=3;
                        break;

                        case 38: /*MOVEI*/
                        gpuregs[gpuset][RN]=gquickread(gpupc)|(gquickread(gpupc+2)<<16);
                        gpupc+=4;
                        cycles-=3;
                        break;

                        case 39: /*LOADB (Rm),Rn*/
//                        if (gpuregs[gpuset][RM]&3) gpub0rk();
                        gpuregs[gpuset][RN]=greadb(gpuregs[gpuset][RM]);
                        cycles-=3;
                        break;

                        case 41: /*LOAD (Rm),Rn*/
                        if (gpuregs[gpuset][RM]&3) gpub0rk();
                        gpuregs[gpuset][RN]=gread(gpuregs[gpuset][RM]+2)|(gread(gpuregs[gpuset][RM])<<16);
                        cycles-=3;
                        break;

                        case 42: /*LOADP (Rm),Rn*/
                        if (gpuregs[gpuset][RM]&3) gpub0rk();
                        gpuregs[gpuset][RN]=gread(gpuregs[gpuset][RM]+2)|(gread(gpuregs[gpuset][RM])<<16);
                        gpuhidata=gread(gpuregs[gpuset][RM]+2)|(gread(gpuregs[gpuset][RM])<<16);
                        cycles-=5;
                        break;

                        case 45: /*STOREB Rn,(Rm)*/
                        gwriteb(gpuregs[gpuset][RM],gpuregs[gpuset][RN]&0xFF);
                        cycles-=3;
                        break;

                        case 47: /*STORE Rn,(Rm)*/
                        if (gpuregs[gpuset][RM]&3) gpub0rk();
                        gwrite(gpuregs[gpuset][RM]+2,gpuregs[gpuset][RN]);
                        gwrite(gpuregs[gpuset][RM],gpuregs[gpuset][RN]>>16);
                        cycles-=3;
                        break;

                        case 48: /*STOREW Rn,(Rm)*/
                        if (gpuregs[gpuset][RM]&1) gpub0rk();
                        gwrite(gpuregs[gpuset][RM],gpuregs[gpuset][RN]);
                        cycles-=3;
                        break;

                        case 51: /*MOVE PC,Rn*/
                        gpupc=gpuregs[gpuset][RN];
                        cycles-=3;
                        break;

                        case 52: /*JUMP cc,(Rn)*/
                        if ((!C0 || !ZFLAG) &&
                            (!C1 || ZFLAG) &&
                            (!C2 || ((C4)?!NFLAG:!CFLAG)) &&
                            (!C3 || ((C4)?NFLAG:CFLAG)))
                        {
                                nextpc=gpuregs[gpuset][RN];
//                                printf("C were %i next %06X\n",(opcode>>5)&31,nextpc);
//                                if (nextpc==0xF032B6) output=1;
                                gpujumpin=2;
                        }
                        cycles-=3;
                        break;

//#if 0
                        case 53: /*JR cc,(Rn)*/
//                        printf("JR\n");
                        if ((!C0 || !ZFLAG) &&
                            (!C1 || ZFLAG) &&
                            (!C2 || ((C4)?!NFLAG:!CFLAG)) &&
                            (!C3 || ((C4)?NFLAG:CFLAG)))
                        {
                                nextpc=gpupc;
                                if (RN&0x10) nextpc-=RN<<1;
                                else         nextpc+=RN<<1;
//                                printf("C were %i next %06X\n",(opcode>>5)&31,nextpc);
                                gpujumpin=2;
                        }
                        cycles-=3;
//                        output=1;
                        break;
//#endif

                        case 57: /*NOP*/
                        cycles-=3;
                        break;

                        default:
                        printf("Bad GPU opcode %i %04X\n",opcode>>10,opcode);
                        dumpgpuregs();
                        exit(-1);
                }
                if (!gpurun) return;
                if (gpujumpin)
                {
                        gpujumpin--;
                        if (!gpujumpin) gpupc=nextpc;
                }
                if (output)
                {
                        printf("%06X %i %i %i ",gpupc,opcode>>10,RM,RN);
                        if ((opcode>>10)==38)     printf("%08X\n",gpuregs[gpuset][RN]);
                        else if ((opcode>>10)==39) printf("%08X\n",gpuregs[gpuset][RM]);
                        else                      printf("\n");
                }
        }
}
