#include <stdio.h>
#include "starcpu.h"

void dumpram();

int dead=0;
unsigned short dspram[0x1000],gpuram[0x800];
unsigned long ram[0x80000],rom[0x100000];
unsigned long tempram[32];
unsigned long long *raml;
unsigned short *ramw;
unsigned long long *roml;
unsigned short *romw;

unsigned readgpu(unsigned addr);
unsigned readjerry(unsigned addr);
unsigned readtom(unsigned addr);
unsigned readblitter(unsigned addr);

void writetom(unsigned addr, unsigned val);
void writejerry(unsigned addr, unsigned val);
void writegpu(unsigned addr, unsigned val);
void writedsp(unsigned addr, unsigned val);
void writeobj(unsigned addr, unsigned val);
void writeclut(unsigned addr, unsigned val);
void writevideo(unsigned addr, unsigned val);
void writeblitter(unsigned addr, unsigned val);

void genwriteb(unsigned addr, unsigned val)
{
        printf("Writing %02X to %08X\n",val,addr);
        exit(-1);
}

void genwritew(unsigned addr, unsigned val)
{
        if ((addr==0x800000)||(addr==0x800002)) return;
        if ((addr==0x802000)||(addr==0x802002)) return;
        printf("Writing %04X to %08X\n",val,addr);
        s68000releaseTimeslice();
        dead=1;
}

unsigned genreadb(unsigned addr)
{
        printf("Reading %08X\n",addr);
        exit(-1);
}

unsigned genreadw(unsigned addr)
{
        if ((addr==0x800000)||(addr==0x800002)) return 0xFFFF;
        if ((addr==0x802000)||(addr==0x802002)) return 0xFFFF;
        printf("Reading word %08X\n",addr);
        s68000releaseTimeslice();
        dead=1;
}

void blankwrite(unsigned addr, unsigned val)
{
}

struct STARSCREAM_PROGRAMREGION jagprg[]=
{
        {0x000000,0x1FFFFF,(unsigned)ram},
        {0x800000,0xBFFFFF,(unsigned)rom-0x800000},
        {-1,-1,NULL}
};

struct STARSCREAM_DATAREGION jagreadb[]=
{
        {0x000000,0x1FFFFF,NULL,ram},
        {0x200000,0x7FFFFF,genreadb,NULL},
        {0x800000,0xBFFFFF,NULL,rom},
        {0xC00000,0xFFFFFF,genreadb,NULL},
        {-1,-1,NULL}
};

struct STARSCREAM_DATAREGION jagreadw[]=
{
        {0x000000,0x1FFFFF,NULL,ram},
        {0x200000,0x7FFFFF,genreadw,NULL},
        {0x800000,0xBFFFFF,NULL,rom},
        {0xC00000,0xF000DF,genreadw,NULL},
        {0xF000E0,0xF000E3,readtom,NULL},
        {0xF000E4,0xF020FF,genreadw,NULL},
        {0xF02100,0xF0211F,readgpu,NULL},
        {0xF02120,0xF021FF,genreadw,NULL},
        {0xF02200,0xF0229F,readblitter,NULL},
        {0xF022A0,0xF02FFF,genreadw,NULL},
        {0xF03000,0xF03FFF,NULL,gpuram},
        {0xF04000,0xF0FFFF,genreadw,NULL},
        {0xF10000,0xF1A0FF,readjerry,NULL},
        {0xF1A100,0xFFFFFF,genreadw,NULL},
        {-1,-1,NULL}
};

struct STARSCREAM_DATAREGION jagwriteb[]=
{
        {0x000000,0x1FFFFF,NULL,ram},
        {0x200000,0xFFFFFF,genwriteb,NULL},
        {-1,-1,NULL}
};

struct STARSCREAM_DATAREGION jagwritew[]=
{
        {0x000000,0x1FFFFF,NULL,ram},
        {0x200000,0xEFFFFF,genwritew,NULL},
        {0xF00000,0xF00003,writetom,NULL},
        {0xF00004,0xF0000F,genwritew,NULL},
        {0xF00010,0xF00027,writeobj,NULL},
        {0xF00028,0xF000DF,writevideo,NULL},
        {0xF000E0,0xF000E3,writetom,NULL},
        {0xF000E4,0xF003FF,genwritew,NULL},
        {0xF00400,0xF007FF,writeclut,NULL},
        {0xF00800,0xF020FF,genwritew,NULL},
        {0xF02100,0xF0211F,writegpu,NULL},
        {0xF02120,0xF021FF,genwritew,NULL},
        {0xF02200,0xF0229F,writeblitter,NULL},
        {0xF022A0,0xF02FFF,genwritew,NULL},
        {0xF03000,0xF03FFF,NULL,gpuram},
        {0xF04000,0xF0FFFF,genwritew,NULL},
        {0xF10000,0xF1A0FF,writejerry,NULL},
        {0xF1A100,0xF1A11F,writedsp,NULL},
        {0xF1A120,0xF1A147,genwritew,NULL},
        {0xF1A148,0xF1A157,writejerry,NULL},
        {0xF1A158,0xF1AFFF,genwritew,NULL},
        {0xF1B000,0xF1CFFF,NULL,dspram},
        {0xF1D000,0xFFFFFF,genwritew,NULL},
        {-1,-1,NULL}
};

void changevectors(int i)
{
        printf("Change vectors %i\n",i);
        exit(-1);
}

void init68k()
{
        s68000init();
        s68000context.s_fetch = jagprg;
        s68000context.u_fetch = jagprg;
        s68000context.s_readbyte  = jagreadb;
        s68000context.u_readbyte  = jagreadb;
        s68000context.s_readword  = jagreadw;
        s68000context.u_readword  = jagreadw;
        s68000context.s_writebyte = jagwriteb;
        s68000context.u_writebyte = jagwriteb;
        s68000context.s_writeword = jagwritew;
        s68000context.u_writeword = jagwritew;
        s68000reset();
        printf("PC am %06X\n",s68000context.pc);
        s68000context.pc=0x802000;//0x20000;//0x4000;
//        s68000context.areg[7]=0x1FFFFE;
//        s68000context.asp=0x1EFFFE;
        atexit(dumpram);
}

void loadrom()
{
        int c=0;
        FILE *f=fopen("test.jag","rb");
        roml=romw=rom;
        raml=ramw=ram;
//        c=0x10000;
        while (!feof(f))
        {
                romw[c]=getc(f)<<8;
                romw[c]|=getc(f);
                c++;
        }
//        fread(ram+0x1000,256*1024,1,f);
        fclose(f);
}

void dumpram()
{
        FILE *f=fopen("ram.dmp","wb");
        fwrite(ram,0x200000,1,f);
        fclose(f);
        f=fopen("dsp.dmp","wb");
        fwrite(dspram,0x2000,1,f);
        fclose(f);
        f=fopen("gpu.dmp","wb");
        fwrite(gpuram,0x1000,1,f);
        fclose(f);
}

void run68k()
{
        unsigned long temp=s68000exec(221600);
        if (dead)
        {
                dump68kregs();
                exit(-1);
        }
        if (temp!=0x80000000)
        {
                printf("Bad result %08X\n",temp);
                exit(-1);
        }
}

void dump68kregs()
{
        printf("D0=%08X D4=%08X A0=%08X A4=%08X\n",s68000context.dreg[0],s68000context.dreg[4],s68000context.areg[0],s68000context.areg[4]);
        printf("D1=%08X D5=%08X A1=%08X A5=%08X\n",s68000context.dreg[1],s68000context.dreg[5],s68000context.areg[1],s68000context.areg[5]);
        printf("D2=%08X D6=%08X A2=%08X A6=%08X\n",s68000context.dreg[2],s68000context.dreg[6],s68000context.areg[2],s68000context.areg[6]);
        printf("D3=%08X D7=%08X A3=%08X A7=%08X\n",s68000context.dreg[3],s68000context.dreg[7],s68000context.areg[3],s68000context.areg[7]);
        printf("PC=%08X SR=%08X\n",s68000context.pc,s68000context.sr);
}
